/*
 * Author:  Konrad Rzeszutek <conradus@mindless.com>
 * Date:    Oct 19th, 1999
 * File:    Editor3D/SmartCanvas.java
 *
 * Notes:
 * 
 *   This object keeps track off all objects relating to the 
 *  graphical part of the canvas. It also handles the rotation,
 *  shearing, scaling, transforming, combinations of these, and
 *  the manipulation of the WC (World Coordinate system).
 *
 *  All these operations are solely based on matrix multiplications.
 * If you don't know matrixes, you better find a Linear Algebra book.
 *
 * Since this class is an Canvas, it has all the operations regarding
 * drawing, putting new figures, and manipulating the object.
 *
 */
package Editor3D;

import java.awt.*;
import java.util.*;
import java.io.*;


public class SmartCanvas extends Canvas {

    // Used for drawing.
    private Dimension OffDim;
    private Image OffImg;
    private Graphics OffG;
    
    private ViewPort viewPort;    // Our holder for Umin, Vmin, Umax, Vmax, VRP, COP....
    private Vector faceList;      // The list of all the faces.
    private Matrix m;             // A matrix for computing.
    private Matrix mTransform;    // The transformation matrix
    private Matrix mI;            // The intensity matrix.
    private boolean debug = false;// Display matrix or not?
    private boolean perspective = true; // Be in perspective or not?
    private boolean backface = false;
    private boolean light = false;
    
    private Vector vectorList;    // List of _all_ points.


    public final static int X_MAX = 400;
    public final static int Y_MAX = 300;
    
    private final static int X_POS = 0;
    private final static int Y_POS = 1;
    private final static int Z_POS = 2;
      
    /*
     * Initailizes the SmartCanvas
     *
     * require:
     *   vi instanceof Editor3D
     */
    public SmartCanvas () {

	this.vectorList = new Vector(0);
	this.faceList = new Vector(0);
	this.viewPort = new ViewPort();
	OffG = null;
	this.m = new Matrix();
	this.mTransform = new Matrix();
	this.mI = new Matrix(3,3);
	mI.m[0][0] = 1; mI.m[1][0] = 8; mI.m[2][0] = 6; 
	mI.m[0][1] = 9; mI.m[1][1] = 4; mI.m[2][1] = 2;
	mI.m[0][2] = 5; mI.m[1][2] = 3; mI.m[2][2] = 7;	
	
    }
    
    /*
     * Read in a new file which will give us the faces, viewport
     *
     * require:
     *   file instanceof String.
     *
     */
    public void newObject(String file) {
    	
	FileObject objFile = new FileObject(file);
	// Get the right data.
	faceList = objFile.getFaceList();
	viewPort = objFile.getViewPort();
	vectorList = objFile.getVectorList(); // All the points.
	recalcMatrix(); // Calculate the view
    }
    

    /*
     * Rotate the object around origin of the canvas.
     *
     * require:
     *   r.length() == 3
     *
     */
    public void rotateObject(double r[]) {
    	
    // Design these huge nice rotation matrix.
	/*
		
	      [ cosZ -sinZ  0 0 ]	[ cosY   0  sinY  0 ]        [ 1  0     0      0 ]     [ x ]
	Rz =  [ sinZ  cosZ  0 0 ]  Ry = [ 0      1  0     0 ]   Rx = [ 0  cosX  -sinX  0 ]  P =[ y ]
	      [ 0     0     1 0 ]       [ -sinY  0  cosY  0 ]        [ 0  sinX  cosX   0 ]     [ z ]
	      [ 0     0     0 1 ]       [ 0      0  0     1 ]        [ 0  0     0      1 ]     [ w ]
	      
	      
	Where, Z, y, and Y are angles. P is a point with x,y,z points.
	
	Since our final matrix is suppose to be: M = Rz*Ry*Rz,
	We can multiply each of the matrices togther, getting:
	
	 Q = Rz*Ry*Rz
	 Pnew = M*Pold
	*/ 
				
	m = new Matrix();
	//	System.out.println(r[0]+","+r[1]+","+r[2]);

	// WARNING: all the numbers entered by the user must be subtraced from 360, so it
	// will rotate the way Dr. Korthright program does it.
	if (r[Editor3D.X_POS] != 0.0) { // User entered something in.
	    double calRxCos = Math.cos(Math.PI* (r[Editor3D.X_POS])/180);
	    double calRxSin = Math.sin(Math.PI* (r[Editor3D.X_POS])/180);
	    m.m[1][1] = m.m[2][2] = calRxCos;
	    m.m[2][1] = calRxSin;
	    m.m[1][2] = - calRxSin;
	    
	}
	if (r[Editor3D.Y_POS] != 0.0) { 
	    double calRyCos = Math.cos(Math.PI * (r[Editor3D.Y_POS])/180);
	    double calRySin = Math.sin(Math.PI * (r[Editor3D.Y_POS])/180);
	    Matrix mRy = new Matrix();
	    mRy.m[0][0] = mRy.m[2][2] = calRyCos;
	    mRy.m[0][2] = calRySin;	    
	    mRy.m[2][0] = -calRySin;
	    
	    m.multiply(mRy,m); // Multiply Rx*Ry
	}
	if (r[Editor3D.Z_POS] != 0.0) { 
	    double calRzCos = Math.cos(Math.PI * (r[Editor3D.Z_POS])/180);
	    double calRzSin = Math.sin(Math.PI * (r[Editor3D.Z_POS])/180);
	    Matrix mRz = new Matrix();
	    mRz.m[0][0] = mRz.m[1][1] = calRzCos;
	    mRz.m[1][0] = calRzSin;	    
	    mRz.m[0][1] = -calRzSin;	    	    
	    m.multiply(mRz,m);	 // Multiply (Rx*Ry)*Rz
	}
	redoCanvas();
    
    }

    /*
     * Scale the object regarding the origin.
     *
     * require:
     *   scale.length() == 3
     */    
    public void scaleObject(double scale[]) {
    
	m = new Matrix();
	
	m.m[0][0] = scale[Editor3D.X_POS];
	m.m[1][1] = scale[Editor3D.Y_POS];
	m.m[2][2] = scale[Editor3D.Z_POS];
	
	redoCanvas();
    }
    
    /*
     * Translate the object by the specified amount from its current
     * position by the specified amount.
     *
     * require:
     *  transform.length() == 3
     */
    public void translateObject(double transform[]) {
    
	m = new Matrix();
	
	m.m[0][3] = transform[Editor3D.X_POS];
	m.m[1][3] = transform[Editor3D.Y_POS];
	m.m[2][3] = transform[Editor3D.Z_POS];

	redoCanvas();
    }
    
    /*
     * Shear the object. This method will shear in all directions
     * as specified by the passed array.
     *
     * require:
     *  shear.length() == 3
     *
     */
    public void shearObject(double[] shear ) {

	Matrix mSh = new Matrix();
	m = new Matrix();
    	// it all depends on which values we have.
	if ( (shear[X_POS] != 0.0) && (shear[Y_POS] != 0.0)) {
	    // XY shear
	    mSh.m[0][2] = shear[X_POS];
	    mSh.m[1][2] = shear[Y_POS];
	}
	else if ((shear[Y_POS] != 0.0) && (shear[Z_POS] != 0.00)){
	    // YZ shear
	    mSh.m[1][0] = shear[Y_POS];
	    mSh.m[2][0] = shear[Z_POS];
	}
	else if ( (shear[X_POS] != 0.0) && (shear[Z_POS] != 0.0)) {
	    // XZ
	    mSh.m[1][1] = shear[X_POS];
	    mSh.m[2][1] = shear[Z_POS];
	}
	// Everything failed.

	else if ((shear[X_POS] != 0.0) && (shear[Y_POS] == 0.00) && (shear[Z_POS] == 0.00)) {
	    // X-shear
	    mSh.m[0][1] = shear[X_POS];
	    mSh.m[2][1] = shear[Z_POS];
	}
	else if ((shear[X_POS] == 0.0) && (shear[Y_POS] != 0.00) && (shear[Z_POS] == 0.00)) {
	    // Y-shear
	    mSh.m[1][0] = shear[Y_POS];
	    mSh.m[2][0] = shear[Z_POS];
	}
	else if ((shear[X_POS] == 0.0) && (shear[Y_POS] == 0.00) && (shear[Z_POS] != 0.00)) {
	    // Z-shear
	    mSh.m[1][0] = shear[X_POS];
	    mSh.m[2][0] = shear[Z_POS];
	}
	m = mSh;
	redoCanvas();
    }
    /*
     * Translate object to the negated specified place (translate[]),
     * then rotate (rotate[]) and move back to the translated place.
     * Essentially this rotates the object based on the specified point--
     * translate[].
     *
     * require:
     *  shear.length() == 3 && translate.length() == 3
     *
     */
    public void trtObject(double[] transform,double[] r) {    

    	m = new Matrix();
	
	m.m[0][3] = transform[Editor3D.X_POS];
	m.m[1][3] = transform[Editor3D.Y_POS];
	m.m[2][3] = transform[Editor3D.Z_POS];


	if (r[Editor3D.X_POS] != 0.0) { // User entered something in.
	    double calRxCos = Math.cos(Math.PI* (r[Editor3D.X_POS])/180);
	    double calRxSin = Math.sin(Math.PI* (r[Editor3D.X_POS])/180);
	    Matrix mRz = new Matrix();
	    mRz.m[1][1] = m.m[2][2] = calRxCos;
	    mRz.m[2][1] = calRxSin;
	    mRz.m[1][2] = - calRxSin;
	    m.multiply(mRz,m);
	    
	}
	if (r[Editor3D.Y_POS] != 0.0) { 
	    double calRyCos = Math.cos(Math.PI * (r[Editor3D.Y_POS])/180);
	    double calRySin = Math.sin(Math.PI * (r[Editor3D.Y_POS])/180);
	    Matrix mRy = new Matrix();
	    mRy.m[0][0] = mRy.m[2][2] = calRyCos;
	    mRy.m[0][2] = calRySin;	    
	    mRy.m[2][0] = -calRySin;	    
	    m.multiply(mRy,m); // Multiply Rx*Ry
	}
	if (r[Editor3D.Z_POS] != 0.0) { 
	    double calRzCos = Math.cos(Math.PI * (r[Editor3D.Z_POS])/180);
	    double calRzSin = Math.sin(Math.PI * (r[Editor3D.Z_POS])/180);
	    Matrix mRz = new Matrix();
	    mRz.m[0][0] = mRz.m[1][1] = calRzCos;
	    mRz.m[1][0] = calRzSin;	    
	    mRz.m[0][1] = -calRzSin;	    	    
	    m.multiply(mRz,m);	 // Multiply (Rx*Ry)*Rz
	}
	
	Matrix mTr = new Matrix();
	
	mTr.m[0][3] = -transform[Editor3D.X_POS];
	mTr.m[1][3] = -transform[Editor3D.Y_POS];
    	mTr.m[2][3] = -transform[Editor3D.Z_POS];
	
	m.multiply(mTr,m); // Multiply Rx*Ry
	        
	redoCanvas();
    }
    /* 
     * Scale the object based on the specified orgin supplied by the
     * translate[] parameter.
     *
     * require: 
     *  scale.length() == 3 && translate.length() == 3
     *
     */
    public void tstObject(double[] translate,double[] scale) {
    
	m = new Matrix();
	Matrix mT = new Matrix();
	mT.m[0][3] = translate[Editor3D.X_POS];
	mT.m[1][3] = translate[Editor3D.X_POS];
	mT.m[2][3] = translate[Editor3D.X_POS];
	
	Matrix mS = new Matrix();
	mS.m[0][0] = scale[Editor3D.X_POS];
	mS.m[1][1] = scale[Editor3D.Y_POS];
	mS.m[2][2] = scale[Editor3D.Z_POS];
	
	m.multiply(mS, mT);

	mT = new Matrix();
	mT.m[0][3] = -translate[Editor3D.X_POS];
	mT.m[1][3] = -translate[Editor3D.X_POS];
	mT.m[2][3] = -translate[Editor3D.X_POS];

	m.multiply(mT, m);
	
	redoCanvas();
	
    }
    
     /* 
     * Shear the object based on the specified orgin supplied by the
     * translate[] parameter.
     *
     * require:
     *  shear.length() == 3 && translate.length() == 3
     *
     */
    public void tshtObjet(double[] translate,double[] shear) {
    
	m = new Matrix();
	Matrix mSh = new Matrix();
    	Matrix mT = new Matrix();

	if (translate != null) {
	    mT.m[0][3] = translate[Editor3D.X_POS];
	    mT.m[1][3] = translate[Editor3D.X_POS];
	    mT.m[2][3] = translate[Editor3D.X_POS];
	}
	// it all depends on which values we have.
	if ( (shear[X_POS] != 0.0) && (shear[Y_POS] != 0.0)) {
	    // XY shear
	    mSh.m[0][2] = shear[X_POS];
	    mSh.m[1][2] = shear[Y_POS];
	}
	else if ((shear[Y_POS] != 0.0) && (shear[Z_POS] != 0.00)){
	    // YZ shear
	    mSh.m[1][0] = shear[Y_POS];
	    mSh.m[2][0] = shear[Z_POS];
	}
	else if ( (shear[X_POS] != 0.0) && (shear[Z_POS] != 0.0)) {
	    // XZ
	    mSh.m[1][1] = shear[X_POS];
	    mSh.m[2][1] = shear[Z_POS];
	}
	// Everything failed.

	else if ((shear[X_POS] != 0.0) && (shear[Y_POS] == 0.00) && (shear[Z_POS] == 0.00)) {
	    // X-shear
	    mSh.m[0][1] = shear[X_POS];
	    mSh.m[2][1] = shear[Z_POS];
	}
	else if ((shear[X_POS] == 0.0) && (shear[Y_POS] != 0.00) && (shear[Z_POS] == 0.00)) {
	    // Y-shear
	    mSh.m[1][0] = shear[Y_POS];
	    mSh.m[2][0] = shear[Z_POS];
	}
	else if ((shear[X_POS] == 0.0) && (shear[Y_POS] == 0.00) && (shear[Z_POS] != 0.00)) {
	    // Z-shear
	    mSh.m[1][0] = shear[X_POS];
	    mSh.m[2][0] = shear[Z_POS];
	}
	
	m.multiply(mSh, mT);


	mT = new Matrix();
	if (translate != null) {
	    mT.m[0][3] = -translate[Editor3D.X_POS];
	    mT.m[1][3] = -translate[Editor3D.X_POS];
	    mT.m[2][3] = -translate[Editor3D.X_POS];
	}
	m.multiply(mT, m);
	
	redoCanvas();
	
    }

    /*
     * Gives the current Umin and Umax variables.
     *
     * return:
     *  length() == 2
     */
    public double[] getU () {
    
	return viewPort.u;
    }
    
    /*
     * Gives the current Vmin and Vmax variables.
     *
     * return:
     *  length() == 2
     */
    
    public double[] getV () {
    
	return viewPort.v;
    }
    
    // Gets various info:
    public VectorObject getVRP() {
    
	return viewPort.vecVRP;
    }
    
    public VectorObject getVPN() {
    
	return viewPort.vecVPN;
    }
    
    public VectorObject getVUP() {
    
	return viewPort.vecVUP;
    }
    public VectorObject getCOP() {
    
	return viewPort.vecCOP;
    }

    public VectorObject getLight() {

	return viewPort.vecLight;
    }
    /*
     * Resets all the modifications done upon the object.
     *
     */
    public void reset () {

       	for (int a = 0; a < faceList.size();a++)
	    ((FaceObject)(faceList.elementAt(a))).reset();
    
	viewPort.reset(); // Problem with COP?
	m = new Matrix();
	redoCanvas();
    }
    
    /*
     * Clear the canvas.
     *
     */
    public void clearArea() {
	
	// clear area bounded by rectangle (x1, y1) -- (x2, y2)
    
	OffG.setColor(getBackground());    
	OffG.fillRect(0, 0, OffDim.width-1, OffDim.height-1);
	OffG.setColor(getForeground());
	
    }

    /*
     * Updates the viewpoint information, also draws the object in  the view
     * of new data.
     *
     * require:
     *  u.length() == v.length() == 2;
     */
    public void update (double[] u, double[] v,
			VectorObject vecVRP,
			VectorObject vecVPN,
			VectorObject vecVUP,
			VectorObject vecCOP,
			VectorObject vecLight,
			boolean debug, 
			boolean perspective,
			boolean backface,
			boolean light) {

       	viewPort.u = u;
	viewPort.v = v;
	viewPort.setVRP(vecVRP);
	viewPort.setVPN(vecVPN);
	viewPort.setVUP(vecVUP);
	viewPort.setCOP(vecCOP);
	viewPort.setLight(vecLight);
	this.debug = debug;
	this.perspective = perspective;
	this.backface = backface;
	this.light = light;
	recalcMatrix();
	redoCanvas();
    }
    /*
     * Updates the debug and perspective flag.
     *
     * require:
     *   boolean type
     *
     */
    public void update(boolean debug, boolean perspective, boolean backface, boolean light) {
	
	this.debug = debug;
	this.perspective = perspective;
	this.backface = backface;
	this.light = light;
    }

    /*
     *  Recalculates the matrix used for transforming 3D objects into
     * a 2D world
     */
    private void recalcMatrix() {
	
	mTransform = new Matrix(); // Reinitialize the matrix
	
	Matrix mT1 = new Matrix(); // T1 matrix -- translating VRP to the origin 
	
	mT1.m[0][3] = -viewPort.vecVRP.vec[X_POS];
	mT1.m[1][3] = -viewPort.vecVRP.vec[Y_POS];
	mT1.m[2][3] = -viewPort.vecVRP.vec[Z_POS];

    	if (debug) System.out.println("Translate:"+mT1);
	
	// For next matrix, we need to find u,v, and n.

	VectorObject vecU = VectorObject.unitVector(VectorObject.crossProduct(viewPort.vecVUP, viewPort.vecVPN));
	VectorObject vecV = VectorObject.unitVector(VectorObject.crossProduct(viewPort.vecVPN, vecU));
	VectorObject vecN = VectorObject.unitVector(viewPort.vecVPN);

	// Design the rotation matrix. M = R*T1(-VRP)
	
	Matrix mR = new Matrix(); // R matrix -- rotating VRC such that n axis (VPN) 
	    // becomes the z axis, the u axis becomes the x axis, and the v-axis
	    // becomes the y axis.
	
	mR.m[0][0] = vecU.vec[X_POS];  mR.m[0][1] = vecU.vec[Y_POS]; mR.m[0][2] = vecU.vec[Z_POS]; 
	mR.m[1][0] = vecV.vec[X_POS];  mR.m[1][1] = vecV.vec[Y_POS]; mR.m[1][2] = vecV.vec[Z_POS]; 
	mR.m[2][0] = vecN.vec[X_POS];  mR.m[2][1] = vecN.vec[Y_POS]; mR.m[2][2] = vecN.vec[Z_POS]; 
	
	if (debug) System.out.println("Rotation:"+mR);
	
	mTransform.multiply(mR,mT1);

	Matrix mT2 = new Matrix(); // Translate the COP into the origin
	
	mT2.m[0][3] = -viewPort.vecCOP.vec[X_POS];
	mT2.m[1][3] = -viewPort.vecCOP.vec[Y_POS];
	mT2.m[2][3] = -viewPort.vecCOP.vec[Z_POS];
	
	if (debug) System.out.println("Translate:"+mT2);
	
	mTransform.multiply(mT2,mTransform);

	Matrix mSh = new Matrix(); // Shear such that the direction of projection becomes parallel to the z-axis
	VectorObject vecCW = new VectorObject( (viewPort.u[Editor3D.MIN] + viewPort.u[Editor3D.MAX])/2.0 ,(viewPort.v[Editor3D.MIN] + viewPort.v[Editor3D.MAX])/2.0, 0,1.0);

	vecCW.multiply(mT2); // Multiply T2*WC

	mSh.m[0][2] = vecCW.vec[X_POS]/viewPort.vecCOP.vec[Z_POS];
     	mSh.m[1][2] = vecCW.vec[Y_POS]/viewPort.vecCOP.vec[Z_POS];

	if (debug) System.out.println("Shear:"+mSh);
	mTransform.multiply(mSh,mTransform);

	// Perspective

	if (perspective) {
	    Matrix mP = new Matrix();
	    mP.m[3][2] = -1.0/viewPort.vecCOP.vec[Z_POS];
	    mP.m[3][3] = 0;
	    if (debug) System.out.println("Perspective:"+mP);
	    	mTransform.multiply(mP,mTransform);
	} 
	// Window to viewport

	// For scalling:
	double dUXRatio = ( (X_MAX) /
		      (viewPort.u[Editor3D.MAX] - viewPort.u[Editor3D.MIN]) );

	double dVYRatio = ( (Y_MAX) / 		    
		     (viewPort.v[Editor3D.MAX] - viewPort.v[Editor3D.MIN]) );

	Matrix mWin = new Matrix();
	mWin.m[0][0] = dUXRatio;        
	mWin.m[0][3] = (viewPort.u[Editor3D.MAX]-viewPort.u[Editor3D.MIN])*.5*dUXRatio;
	mWin.m[1][1] = -dVYRatio;
	mWin.m[1][3] = (viewPort.v[Editor3D.MAX]-viewPort.v[Editor3D.MIN])*0.5*dVYRatio;
	    
	if (debug) System.out.println("Window to viewport:"+mWin);
	
	mTransform.multiply(mWin,mTransform);
	
	if (debug) System.out.println("Final View Matrix:"+mTransform);

	// Find the COP in WC, as opposed to to COP in VRC.

    }
 

    /*
     * Re-make the canvas. Compute the new object faces based on
     * the 'm' matrix. Then redraw the canvas.
     *
     */
    public void redoCanvas() {

	DrawLine dl;
	int i,j;
	clearArea();

	// Display the matrix (if required)
	if (debug)
	    System.out.println("Operation matrix:"+m);
	    
	// Do the required computation on the object 

	for (int a = 0; a < vectorList.size();a++) {
	    // Multiply by the operatin matrix, recalculate the WC points (using combination
	    // matrix earlier dereived and homogonize points.
	    ((VectorObject)(vectorList.elementAt(a))).multiply(m);
	    ((VectorObject)(vectorList.elementAt(a))).recalcWC(mTransform);
	    ((VectorObject)(vectorList.elementAt(a))).homogonize(); 
	}


	for (i = 0; i < faceList.size();i++)  {
	    
	    // Go thru each face.
	    if (backface && !light) {
		if (((FaceObject)(faceList.elementAt(i))).backFace(viewPort.vecCOP, viewPort.vecLight))
		    // Calculate the normal.
		    for (j = 0; j < ((FaceObject)(faceList.elementAt(i))).size()-1;j++) 
		    	dl = new DrawLine(OffG,OffDim,			    
					  (int)((FaceObject)(faceList.elementAt(i))).get(j).wc[VectorObject.X_POS],
					  (int)((FaceObject)(faceList.elementAt(i))).get(j).wc[VectorObject.Y_POS],
					  (int)((FaceObject)(faceList.elementAt(i))).get(j+1).wc[VectorObject.X_POS],
					  (int)((FaceObject)(faceList.elementAt(i))).get(j+1).wc[VectorObject.Y_POS]);
	    }				    
	    if (light) {
		if (((FaceObject)(faceList.elementAt(i))).backFace(viewPort.vecCOP, viewPort.vecLight)) {
		    // Let there be light.
		    
		    int illum  = ((FaceObject)(faceList.elementAt(i))).illum;
		    
		    int ints;              // Simple counters
		    int n = ((FaceObject)(faceList.elementAt(i))).size()-1;
		    
		    int ymin, ymax,y,xx;     
		    int[] x;
		    int pt1x, pt1y;
		    int pt0x, pt0y;
		    
		    if (n > 1) { // Better have more then one edge.
			// Lets start with the first object.	
			pt0x = (int) ((FaceObject)faceList.elementAt(i)).get(0).wc[VectorObject.X_POS];
			pt0y = (int) ((FaceObject)faceList.elementAt(i)).get(0).wc[VectorObject.Y_POS];
			
			ymin = ymax = pt0y;
			for (j=0;j<n;j++) { // Go thru the edges and find ymin and ymax.
			    
			    pt0y = (int) ((FaceObject)faceList.elementAt(i)).get(j).wc[VectorObject.Y_POS];		
			    if (pt0y < ymin) 
				ymin = pt0y;
			    if (pt0y > ymax)
				ymax = pt0y;
			}
			
			x = new int[n*n]; // is this enough? N is the amount of edges, so in worst
			// case scenario the points are zigzaging like sine around y and would have n*n X's.
			
			for (y=ymin;y<=ymax;y++) {
			    // find intersections of scan line y with all edges and leave them in 
			    // array X. 
			    ints = 0;
			    
			    for (j=0;j<n;j++) { // for all edges, and for all y's
				pt0x = (int) ((FaceObject)faceList.elementAt( i )).get(j).wc[VectorObject.X_POS];
				pt0y = (int) ((FaceObject)faceList.elementAt( i )).get(j).wc[VectorObject.Y_POS];
				
				pt1x = (int) ((FaceObject)faceList.elementAt( i )).get((j+1) % n).wc[VectorObject.X_POS];
				pt1y = (int) ((FaceObject)faceList.elementAt( i )).get((j+1) % n).wc[VectorObject.Y_POS];
				
				if ( 
				    (((y >= pt0y) && (y <= pt1y)) ||  // Is y between y0 and y1 
				     ((y <= pt0y) && (y >= pt1y))) &&
				    ((pt0y != pt1y)) &&		// Is y not horizontal
				    ((y != ymax)))			// Is y not the highest endpoint
				    {
					x[ints] = (int)( (double)(pt0x) + // We have to convert first to double and then to int.
							 (((double)(y-pt0y)) * ((double)(pt1x - pt0x)) / ((double)(pt1y-pt0y))));
					ints++;			   
				    }
			    }	    
			    if (ints >= 1) {
				// Sort them.
				QuickSort.IntegerQuickSort(x,0, ints-1);
				for (j=0;j<ints;j++) {
				    for (xx=x[j];((xx<=x[j+1]) && ((j+1) < ints));xx++)  {
					if (((xx % 3) >= 0) && ((y % 3) >= 0) )
					    if ((mI.m[(xx % 3)][y % 3]) < illum )
						OffG.drawRect(xx,y,0,0);
				    }
				}	
			    }  
			}
		    }
		    OffG.setColor(Color.white); // In case we desynced on the last Y.
		}
	    }
	    else if (!backface && !light) {
		
		for (j = 0; j < ((FaceObject)(faceList.elementAt(i))).size()-1;j++) 
		    dl = new DrawLine(OffG,OffDim,								  
				      (int)((FaceObject)(faceList.elementAt(i))).get(j).wc[VectorObject.X_POS],
				      (int)((FaceObject)(faceList.elementAt(i))).get(j).wc[VectorObject.Y_POS],
				      (int)((FaceObject)(faceList.elementAt(i))).get(j+1).wc[VectorObject.X_POS],
				      (int)((FaceObject)(faceList.elementAt(i))).get(j+1).wc[VectorObject.Y_POS]); }
	} // end of loop
	
	m = new Matrix();
	repaint();
    }  

  /*
   * Paint the canvas on the supplied g Graphic component.
   */  
  public void paint(Graphics g) {
        
    // make sure offscreen drawing stuff is initialized.  This stuff has to be
    // initialized here because the canvas' size is not yet valid in the
    // constructor

    if (OffG == null) {
      // get size of drawing area
      OffDim = getSize();   
      // create a place to draw offscreen
      OffImg = createImage(OffDim.width, OffDim.height); 
      // get graphics context
      OffG = OffImg.getGraphics();       
      
      // clear offscreen drawing area 
      clearArea();      
    }    
    // draw onscreen from offscreen drawing area
    
    g.drawImage(OffImg, 0, 0, this);
    
  }
    
}
